@@ -11,6 +11,7 @@ from qpd.specs import (
     WindowFunctionSpec,
 )
 from triad.utils.assertion import assert_or_throw
+from qpd.pandas_like import PandasLikeUtils
 
 
 class QPDEngine(ABC):
@@ -26,6 +27,11 @@ class QPDEngine(ABC):
     def assemble_df(self, *args: Any) -> DataFrame:
         return DataFrame(*args)
 
+    @property
+    @abstractmethod
+    def pl_utils(self) -> PandasLikeUtils:
+        raise NotImplementedError
+
     @abstractmethod
     def to_df(self, obj: Any) -> DataFrame:  # pragma: no cover
         raise NotImplementedError
@@ -209,80 +215,29 @@ class QPDEngine(ABC):
 
     def drop_duplicates(self, df: DataFrame) -> DataFrame:
         ndf = self.to_native(df)
-        return self.to_df(self._drop_duplicates(ndf))
+        return self.to_df(self.pl_utils.drop_duplicates(ndf))
 
     def union(self, df1: DataFrame, df2: DataFrame, unique: bool) -> DataFrame:
-        ndf1, ndf2 = self._preprocess_set_op(df1, df2)
-        ndf = ndf1.append(ndf2)
-        if unique:
-            ndf = self._drop_duplicates(ndf)
-        return self.to_df(ndf)
+        ndf1 = self.to_native(df1)
+        ndf2 = self.to_native(df2)
+        return self.to_df(self.pl_utils.union(ndf1, ndf2, unique))
 
     def intersect(self, df1: DataFrame, df2: DataFrame, unique: bool) -> DataFrame:
-        ndf1, ndf2 = self._preprocess_set_op(df1, df2)
-        ndf = ndf1.merge(self._drop_duplicates(ndf2))
-        if unique:
-            ndf = self._drop_duplicates(ndf)
-        return self.to_df(ndf)
+        ndf1 = self.to_native(df1)
+        ndf2 = self.to_native(df2)
+        return self.to_df(self.pl_utils.intersect(ndf1, ndf2, unique))
 
     def except_df(self, df1: DataFrame, df2: DataFrame, unique: bool) -> DataFrame:
-        ndf1, ndf2 = self._preprocess_set_op(df1, df2)
-        ndf2["__anti_indicator__"] = 1
-        ndf = ndf1.merge(ndf2, how="left", on=list(ndf1.columns))
-        ndf = ndf[ndf["__anti_indicator__"].isnull()].drop(
-            ["__anti_indicator__"], axis=1
-        )
-        if unique:
-            ndf = self._drop_duplicates(ndf)
-        return self.to_df(ndf)
+        ndf1 = self.to_native(df1)
+        ndf2 = self.to_native(df2)
+        return self.to_df(self.pl_utils.except_df(ndf1, ndf2, unique))
 
     def join(
         self, df1: DataFrame, df2: DataFrame, join_type: str, on: List[str]
     ) -> DataFrame:
         ndf1 = self.to_native(df1)
         ndf2 = self.to_native(df2)
-        if join_type == "inner":
-            ndf1 = ndf1.dropna(subset=on)
-            ndf2 = ndf2.dropna(subset=on)
-            joined = ndf1.merge(ndf2, how=join_type, on=on)
-        elif join_type == "left_semi":
-            ndf1 = ndf1.dropna(subset=on)
-            ndf2 = self._drop_duplicates(ndf2[on].dropna())
-            joined = ndf1.merge(ndf2, how="inner", on=on)
-        elif join_type == "left_anti":
-            ndf2 = self._drop_duplicates(ndf2[on].dropna())
-            ndf2["__anti_indicator__"] = 1
-            joined = ndf1.merge(ndf2, how="left", on=on)
-            joined = joined[joined["__anti_indicator__"].isnull()].drop(
-                ["__anti_indicator__"], axis=1
-            )
-        elif join_type == "left_outer":
-            ndf2 = ndf2.dropna(subset=on)
-            joined = ndf1.merge(ndf2, how="left", on=on)
-        elif join_type == "right_outer":
-            ndf1 = ndf1.dropna(subset=on)
-            joined = ndf1.merge(ndf2, how="right", on=on)
-        elif join_type == "full_outer":
-            add: List[str] = []
-            for f in on:
-                name = f + "_null"
-                s1 = ndf1[f].isnull().astype(int)
-                ndf1[name] = s1
-                s2 = ndf2[f].isnull().astype(int) * 2
-                ndf2[name] = s2
-                add.append(name)
-            joined = ndf1.merge(ndf2, how="outer", on=on + add).drop(add, axis=1)
-        elif join_type == "cross":
-            assert_or_throw(
-                len(on) == 0, ValueError(f"cross join can't have join keys {on}")
-            )
-            ndf1["__cross_indicator__"] = 1
-            ndf2["__cross_indicator__"] = 1
-            joined = ndf1.merge(ndf2, how="inner", on=["__cross_indicator__"]).drop(
-                ["__cross_indicator__"], axis=1
-            )
-        else:  # pragma: no cover
-            raise NotImplementedError(join_type)
+        joined = self.pl_utils.join(ndf1, ndf2, join_type=join_type, on=on)
         return self.to_df(joined)
 
     def _set_op_result_to_none(self, series: Any, s1: Any, s2: Any) -> Column:
@@ -315,6 +270,3 @@ class QPDEngine(ABC):
             return float("nan")
         else:
             return float(col.native > 0)
-
-    def _drop_duplicates(self, ndf: Any) -> Any:
-        return ndf.drop_duplicates()
