@@ -3,9 +3,7 @@ from typing import Any, Callable, Dict, List, Optional, Tuple
 import numpy as np
 import pandas as pd
 from pandas.core.window.indexers import BaseIndexer
-from triad.utils.assertion import assert_or_throw
-
-from qpd import QPDEngine, run_sql
+from qpd import PandasLikeUtils, QPDEngine, run_sql
 from qpd.dataframe import Column, DataFrame
 from qpd.specs import (
     AggFunctionSpec,
@@ -14,6 +12,7 @@ from qpd.specs import (
     RollingWindowFrame,
     WindowFunctionSpec,
 )
+from triad.utils.assertion import assert_or_throw
 
 
 def run_sql_on_pandas(
@@ -22,7 +21,15 @@ def run_sql_on_pandas(
     return run_sql(QPDPandasEngine(), sql, dfs, ignore_case=ignore_case)
 
 
+class PandasUtils(PandasLikeUtils[pd.DataFrame, pd.Series]):
+    pass
+
+
 class QPDPandasEngine(QPDEngine):
+    @property
+    def pl_utils(self) -> PandasUtils:
+        return PandasUtils()
+
     def to_df(self, obj: Any) -> DataFrame:
         if isinstance(obj, DataFrame):
             return obj
