@@ -6,6 +6,8 @@ class SQLTests(object):
     class Tests(TestsBase):
         def test_basic_select_from(self):
             df = self.make_rand_df(5, a=(int, 2), b=(str, 3), c=(float, 4))
+            self.eq_sqlite("SELECT 1 AS a, 1.5 AS b, 'x' AS c")
+            self.eq_sqlite("SELECT 1+2 AS a, 1.5*3 AS b, 'x' AS c")
             self.eq_sqlite("SELECT * FROM a", a=df)
             self.eq_sqlite("SELECT * FROM a AS x", a=df)
             self.eq_sqlite("SELECT b AS bb, a+1-2*3.0/4 AS cc, x.* FROM a AS x", a=df)
